import json
import random
import time
import requests
import websocket
from typing import List, Callable, Any
from hexbytes import HexBytes
from hubble_exchange.models import (
    Order,
    OrderStatusResponse,
    OrderBookDepthResponse,
    GetPositionsResponse,
    OrderBookDepthUpdateResponse,
    Message,
    WebsocketResponse,
)
from hubble_exchange.order_book import OrderBookClient
from hubble_exchange.utils import (
    get_rpc_endpoint,
    get_websocket_endpoint,
    float_to_scaled_int,
    get_address_from_private_key,
    get_new_salt,
)


class HubbleClient:
    def __init__(self, private_key: str):
        if not private_key:
            raise ValueError("Private key is not set")
        self.trader_address = get_address_from_private_key(private_key)
        if not self.trader_address:
            raise ValueError("Cannot determine trader address from private key")

        self.rpc_endpoint = get_rpc_endpoint()
        self.websocket_endpoint = get_websocket_endpoint()
        self.orderBookClient = OrderBookClient(private_key)

    def get_order_book(self, market: int) -> OrderBookDepthResponse:
        request_body = {
            "jsonrpc": "2.0",
            "id": 1,
            "method": "trading_getTradingOrderBookDepth",
            "params": [market],
        }
        response = requests.post(self.rpc_endpoint, json=request_body)
        response_json = response.json()
        order_book = OrderBookDepthResponse(**response_json["result"])
        return order_book

    def get_margin_and_positions(self) -> GetPositionsResponse:
        request_body = {
            "jsonrpc": "2.0",
            "id": 1,
            "method": "trading_getMarginAndPositions",
            "params": [self.trader_address],
        }
        response = requests.post(self.rpc_endpoint, json=request_body)
        response_json = response.json()
        positions = GetPositionsResponse(**response_json["result"])
        return positions

    def get_order_status(self, order_id: HexBytes) -> OrderStatusResponse:
        request_body = {
            "jsonrpc": "2.0",
            "id": 1,
            "method": "trading_getOrderStatus",
            "params": [order_id.hex()],
        }
        response = requests.post(self.rpc_endpoint, json=request_body)
        response_json = response.json()
        order_status = OrderStatusResponse(**response_json["result"])
        return order_status

    def place_orders(self, orders: List[Order]) -> List[Order]:
        if len(orders) > 75:
            raise ValueError("Cannot place more than 75 orders at once")

        for order in orders:
            if order.amm_index is None:
                raise ValueError("Order AMM index is not set")
            if order.base_asset_quantity is None:
                raise ValueError("Order base asset quantity is not set")
            if order.price is None:
                raise ValueError("Order price is not set")
            if order.reduce_only is None:
                raise ValueError("Order reduce only is not set")

            # trader and salt can be set automatically
            if order.trader in [None, "0x", ""]:
                order.trader = self.trader_address
            if order.salt in [None, 0]:
                order.salt = get_new_salt()

        return self.orderBookClient.place_orders(orders)

    def place_single_order(
        self, market: int, base_asset_quantity: float, price: float, reduce_only: bool
    ) -> Order:
        salt = str(int(time.time())) + str(random.randint(0, 1000))
        salt_int = int(salt)
        order = Order(
            id=None,
            amm_index=market,
            trader=self.trader_address,
            base_asset_quantity=float_to_scaled_int(base_asset_quantity, 18),
            price=float_to_scaled_int(price, 6),
            salt=get_new_salt(),
            reduce_only=reduce_only,
        )
        tx_hash = self.orderBookClient.place_order(order)
        order.id = tx_hash
        return order

    def cancel_orders(self, orders: List[Order]) -> None:
        self.orderBookClient.cancel_orders(orders)

    def cancel_order_by_id(self, order_id: HexBytes) -> None:
        order_status = self.get_order_status(order_id)
        order = Order(
            amm_index=order_status.Symbol,
            trader=self.trader_address,
            base_asset_quantity=float_to_scaled_int(order_status.OrigQty, 18),
            price=float_to_scaled_int(order_status.Price, 6),
            salt=order_status.Salt,
            reduce_only=order_status.ReduceOnly,
        )
        self.cancel_orders([order])

    def subscribe_to_order_book_depth(
        self, market: int, callback: Callable[[websocket.WebSocketApp, OrderBookDepthUpdateResponse], Any]
    ) -> None:

        def on_open(ws):
            msg = {
                "jsonrpc": "2.0",
                "id": 1,
                "method": "trading_subscribe",
                "params": ["streamDepthUpdateForMarket", market]
            }
            ws.send(json.dumps(msg))

        def on_message(ws, message):
            message_json = json.loads(message)
            response = WebsocketResponse(**message_json)
            if response.method and response.method == "trading_subscription":
                response = OrderBookDepthUpdateResponse(
                    T=response.params['result']['T'],
                    symbol=response.params['result']['s'],
                    bids=response.params['result']['b'],
                    asks=response.params['result']['a'],
                )
                callback(ws, response)

        ws = websocket.WebSocketApp(self.websocket_endpoint, on_open=on_open, on_message=on_message)
        ws.run_forever()



import os
import math
import random
import time
from eth_typing import Address
from eth_account import Account


rpc_endpoint = ""
websocket_endpoint = ""


def float_to_scaled_int(val: float, scale: int) -> int:
    big_float = math.floor(val * (10 ** scale))
    return int(big_float)


def get_address_from_private_key(private_key: str) -> Address:
    account = Account.from_key(private_key)
    return account.address

def get_new_salt() -> int:
    return int(str(time.time_ns()) + str(random.randint(0, 10000)))


def get_rpc_endpoint() -> str:
    global rpc_endpoint
    if not rpc_endpoint:
        rpc_host = os.getenv("HUBBLE_RPC_HOST")
        if not rpc_host:
            raise ValueError("HUBBLE_RPC_HOST environment variable not set")
        blockchain_id = os.getenv("HUBBLE_BLOCKCHAIN_ID")
        if not blockchain_id:
            raise ValueError("HUBBLE_BLOCKCHAIN_ID environment variable not set")
        path = f"/ext/bc/{blockchain_id}/rpc"
        rpc_endpoint = f"https://{rpc_host}{path}"
    return rpc_endpoint


def get_websocket_endpoint() -> str:
    global websocket_endpoint
    if not websocket_endpoint:
        rpc_host = os.getenv("HUBBLE_RPC_HOST")
        if not rpc_host:
            raise ValueError("HUBBLE_RPC_HOST environment variable not set")
        blockchain_id = os.getenv("HUBBLE_BLOCKCHAIN_ID")
        if not blockchain_id:
            raise ValueError("HUBBLE_BLOCKCHAIN_ID environment variable not set")
        path = f"/ext/bc/{blockchain_id}/ws"
        websocket_endpoint = f"wss://{rpc_host}{path}"
    return websocket_endpoint



import concurrent.futures
import json
import os

from eth_typing import Address
from hexbytes import HexBytes
from typing import List, Dict, Any
from web3 import Web3
from web3.middleware import geth_poa_middleware
from web3.middleware.cache import construct_simple_cache_middleware

from hubble_exchange.models import Order
from hubble_exchange.utils import get_rpc_endpoint, get_address_from_private_key

MAX_GAS_LIMIT = 15_000_000  # 15 million
OrderBookContractAddress = Address("0x0300000000000000000000000000000000000000")

# read abi from file
dir_ = os.path.dirname(__file__)
with open(f"{dir_}/contract_abis/OrderBook.json", 'r') as abi_file:
    abi_str = abi_file.read()
    ABI = json.loads(abi_str)


class OrderBookClient(object):
    def __init__(self, private_key: str):
        self.rpc_endpoint = get_rpc_endpoint()
        self._private_key = private_key
        self.public_address = get_address_from_private_key(private_key)
        self.nonce = None
        self.chain_id = 321123

        self.web3_client = Web3(Web3.HTTPProvider(self.rpc_endpoint))
        self.web3_client.middleware_onion.inject(geth_poa_middleware, layer=0)

        # cache frequent eth_chainId calls
        cache_chain_id_middleware = construct_simple_cache_middleware()
        self.web3_client.middleware_onion.add(cache_chain_id_middleware, name="cache")

        self.order_book = self.web3_client.eth.contract(address=OrderBookContractAddress, abi=ABI)

    def place_order(self, order: Order) -> HexBytes:
        order_hash = self._get_order_hash(order)

        self._send_orderbook_transaction("placeOrder", [order.to_dict()], {'gas': 200_000})
        return order_hash

    def place_orders(self, orders: List[Order]) -> List[Order]:
        """
        Place multiple orders at once. This is more efficient than placing them one by one.
        """
        place_order_payload = []

        # get order hashes concurrently
        with concurrent.futures.ThreadPoolExecutor(max_workers=100) as executor:
            order_hashes = list(executor.map(self._get_order_hash, orders))

        for order, order_hash_bytes in zip(orders, order_hashes):
            order_hash = HexBytes(order_hash_bytes)
            place_order_payload.append(order.to_dict())
            order.id = order_hash

        self._send_orderbook_transaction("placeOrders", [place_order_payload], {'gas': min(200_000 * len(orders), MAX_GAS_LIMIT)})
        return orders


    def cancel_orders(self, orders: list[Order]) -> None:
        cancel_order_payload = []
        for order in orders:
            cancel_order_payload.append(order.to_dict())
        self._send_orderbook_transaction("cancelOrders", [cancel_order_payload])


    def _get_nonce(self) -> int:
        if self.nonce is None:
            self.nonce = self.web3_client.eth.get_transaction_count(self.public_address)
        else:
            self.nonce += 1
        return self.nonce
    
    def _get_order_hash(self, order: Order) -> HexBytes:
        order_hash_bytes = self.order_book.functions.getOrderHash(order.to_dict()).call()
        return HexBytes(order_hash_bytes)

    def _send_orderbook_transaction(self, method_name: str, args: List[Any], tx_options: Dict = None) -> HexBytes:
        method = getattr(self.order_book.functions, method_name)
        nonce = self._get_nonce()
        tx_params = {
            'from': self.public_address,
            'chainId': self.chain_id,
            'maxFeePerGas': Web3.to_wei(60, 'gwei'),
            'maxPriorityFeePerGas': 0,
            'nonce': nonce,
        }
        if tx_options:
            tx_params.update(tx_options)

        transaction = method(*args).build_transaction(tx_params)
        signed_tx = self.web3_client.eth.account.sign_transaction(transaction, self._private_key)
        tx_hash = self.web3_client.eth.send_raw_transaction(signed_tx.rawTransaction)
        return tx_hash
